import pytest
from unittest.mock import MagicMock
from ibind.client.ibkr_client import IbkrClient

@pytest.fixture
def client():
    # Minimal config for IbkrClient, mock dependencies
    c = IbkrClient(use_oauth=False)
    c.check_auth_status = MagicMock()
    c.stop_tickler = MagicMock()
    c.oauth_init = MagicMock()
    c.oauth_config = MagicMock()
    c.oauth_config.maintain_oauth = True
    c.oauth_config.init_brokerage_session = True
    c.oauth_config.shutdown_oauth = False
    return c


def test_handle_auth_status_healthy(client, caplog):
    ## Arrange
    client.check_auth_status.return_value = True

    ## Act
    with caplog.at_level("WARNING", logger="ibind.client.ibkr_client"):
        assert client.handle_auth_status() is True

    ## Assert
    # No warning should be logged
    assert not any("IBKR connection is not healthy" in r.message for r in caplog.records)
    client.check_auth_status.assert_called_once()
    client.stop_tickler.assert_not_called()
    client.oauth_init.assert_not_called()


def test_handle_auth_status_not_healthy_no_oauth(client, caplog):
    ## Arrange
    client.check_auth_status.return_value = False
    client._use_oauth = False

    ## Act
    with caplog.at_level("WARNING", logger="ibind.client.ibkr_client"):
        assert client.handle_auth_status() is False

    ## Assert
    assert any("IBKR connection is not healthy. Ensure authentication with the Gateway is re-established." in r.message for r in caplog.records)
    client.stop_tickler.assert_not_called()
    client.oauth_init.assert_not_called()


def test_handle_auth_status_not_healthy_oauth_success(client, caplog):
    ## Arrange
    client.check_auth_status.return_value = False
    client._use_oauth = True
    client.stop_tickler.side_effect = None
    client.oauth_init.side_effect = None

    ## Act
    with caplog.at_level("WARNING", logger="ibind.client.ibkr_client"):
        assert client.handle_auth_status() is False

    ## Assert
    assert any("IBKR connection is not healthy. Attempting to re-establish OAuth authentication." in r.message for r in caplog.records)
    client.stop_tickler.assert_called_once_with(15)
    client.oauth_init.assert_called_once_with(maintain_oauth=True, init_brokerage_session=True)


import json
import logging
from threading import Thread
from typing import Optional
from unittest import TestCase
from unittest.mock import MagicMock, patch, call

import requests

from ibind import Result
from ibind.client.ibkr_client import IbkrClient
from ibind.client.ibkr_ws_client import IbkrWsClient, IbkrSubscriptionProcessor, IbkrWsKey
from ibind.support.logs import project_logger
from test.integration.base.websocketapp_mock import create_wsa_mock, init_wsa_mock
from test_utils import RaiseLogsContext, SafeAssertLogs


class TestPreprocessRawMessage(TestCase):
    def setUp(self):
        self.url = 'wss://localhost:5000/v1/api/ws'

        self.ws_client = IbkrWsClient(
            url=self.url,
            ibkr_client=None,
            account_id=None,
            subscription_processor_class=lambda: None,
        )

    def test_preprocess_with_well_formed_message(self):
        raw_message = json.dumps({'topic': 'actABC', 'args': {'key': 'value'}})
        expected_result = (
            {'topic': 'actABC', 'args': {'key': 'value'}},  # message
            'actABC',  # topic
            {'key': 'value'},  # data
            'a',  # subscribed
            'ctABC',  # channel
        )
        self.assertEqual(self.ws_client._preprocess_raw_message(raw_message), expected_result)

    def test_preprocess_with_unsubscribed_message(self):
        raw_message = json.dumps({'message': 'Unsubscribed'})
        expected_result = ({'message': 'Unsubscribed'}, None, None, None, None)
        self.assertEqual(self.ws_client._preprocess_raw_message(raw_message), expected_result)


class TestIbkrWsClient(TestCase):
    # Assuming IbkrWsClient is the class containing preprocess_raw_message

    def setUp(self):
        # Assuming similar initialization parameters as in WsClient
        self.url = 'wss://localhost:5000/v1/api/ws'
        self.max_reconnect_attempts = 4
        self.max_ping_interval = 38

        self.url_rest = 'https://localhost:5000'
        self.account_id = 'TEST_ACCOUNT_ID'
        self.timeout = 8
        self.max_retries = 4
        self.subscription_retries = 3
        self.client = MagicMock(
            spec=IbkrClient(
                url=self.url_rest,
                account_id=self.account_id,
                timeout=self.timeout,
                max_retries=self.max_retries,
            )
        )

        self.client.tickle.return_value.data = {'session': 'TEST_COOKIE'}

        self.SubscriptionProcessorClass = IbkrSubscriptionProcessor

        # Initialize the IbkrWsClient
        self.ws_client = IbkrWsClient(
            url=self.url,
            ibkr_client=self.client,
            account_id=self.account_id,
            subscription_processor_class=self.SubscriptionProcessorClass,
            subscription_retries=self.subscription_retries,
            subscription_timeout=0.01,
            cacert=False,
            timeout=0.01,
            max_connection_attempts=self.max_reconnect_attempts,
            max_ping_interval=self.max_ping_interval,
        )

        self.wsa_mock = create_wsa_mock()
        self.thread_mock = MagicMock(spec=Thread)
        self.thread_mock.start.side_effect = lambda: self.ws_client._run_websocket(self.wsa_mock)

        self.conid = 265598
        self.update_time = 5678765456

    def run_in_test_context(self, fn, expected_errors: list[str] = None, expect_logs: bool = True):
        with patch('ibind.base.ws_client.WebSocketApp', side_effect=lambda *args, **kwargs: init_wsa_mock(self.wsa_mock, *args, **kwargs)), \
                patch('ibind.base.ws_client.Thread', return_value=self.thread_mock) as new_thread_mock, \
                SafeAssertLogs(self, 'ibind', level='DEBUG', logger_level='DEBUG', no_logs=not expect_logs) as cm, \
                RaiseLogsContext(self, 'ibind', level='WARNING', expected_errors=expected_errors):  # fmt: skip
            ws_client_logger = project_logger('ws_client')
            old_level = ws_client_logger.getEffectiveLevel()
            ws_client_logger.setLevel(logging.WARNING)

            self.new_thread_mock = new_thread_mock
            try:
                rv = fn()
            except:
                raise
            finally:
                ws_client_logger.setLevel(old_level)

        return cm, rv

    def _send_payload(self, payload: dict, expected_errors: list[str] = None, expect_logs: bool = True):
        def run():
            success = self.ws_client.start()
            raw_payload = json.dumps(payload)
            self.ws_client.send(raw_payload)
            self.ws_client.shutdown()
            return success

        return self.run_in_test_context(run, expected_errors=expected_errors, expect_logs=expect_logs)

    def _subscribe(self, request: dict, response: Optional[dict], expected_errors: list[str] = None, expect_logs: bool = True):
        def run():
            def override_on_message(wsa_mock: MagicMock, message: str):
                if response is None:
                    return
                raw_message = json.dumps(response)
                wsa_mock.__on_message__(wsa_mock, raw_message)

            self.ws_client.start()
            self.wsa_mock._on_message.side_effect = override_on_message
            rv = self.ws_client.subscribe(
                **{'channel': request.get('channel'), 'data': request.get('data'), 'needs_confirmation': request.get('needs_confirmation')}
            )
            self.ws_client.unsubscribe(
                **{'channel': request.get('channel'), 'data': request.get('data'), 'needs_confirmation': request.get('confirms_unsubscription')}
            )
            self.ws_client.shutdown()
            return rv

        return self.run_in_test_context(run, expected_errors=expected_errors, expect_logs=expect_logs)

    def test_on_message_system_heartbeat(self):
        hb = 12345678
        cm, success = self._send_payload({'topic': 'system', 'hb': hb}, expect_logs=False)
        # print("\n".join([r.msg for r in cm.records]))
        self.assertEqual(self.ws_client._last_heartbeat, hb)

    def test_on_message_act_account_mismatch(self):
        message_data = {'topic': 'act', 'args': {'accounts': ['OTHER_ACCOUNT_ID']}}
        expected_errors = ["IbkrWsClient: Account ID mismatch: expected=TEST_ACCOUNT_ID, received=['OTHER_ACCOUNT_ID']"]

        cm, success = self._send_payload(message_data, expected_errors=expected_errors)
        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def test_on_message_blt(self):
        bulletin_message = {'topic': 'blt', 'args': {'bulletin_key': 'some_info'}}

        with patch.object(self.ws_client, '_handle_bulletin', MagicMock()) as mock_handle_bulletin:
            cm, success = self._send_payload(bulletin_message, expect_logs=False)
            mock_handle_bulletin.assert_called_once_with(bulletin_message)

    def test_on_message_sts_unauthenticated(self):
        message_data = {'topic': 'sts', 'args': {'authenticated': False}}
        session_id = 6545676

        expected_errors = ["IbkrWsClient: Status unauthenticated: {'authenticated': False}", 'IbkrWsClient: Not authenticated, closing WebSocketApp']

        response_mock = MagicMock(spec=requests.Response)
        response_mock.status_code = 200
        response_mock.json.return_value = {'session': session_id, 'data_to_be_ignored': '1234'}

        self.client.tickle.return_value = Result(data=response_mock.json.return_value)

        with patch('ibind.base.rest_client.requests') as requests_mock:
            requests_mock.request.return_value = response_mock
            cm, success = self._send_payload(message_data, expected_errors=expected_errors)

        self.assertEqual(expected_errors, [r.msg for r in cm.records])
        self.assertFalse(self.ws_client._authenticated)

    def test_on_message_sts_authenticated(self):
        message_data = {'topic': 'sts', 'args': {'authenticated': True}}
        cm, success = self._send_payload(message_data, expect_logs=False)

    def test_on_message_error(self):
        message_data = {'topic': 'error', 'args': {'error_key': 'error_details'}}
        expected_errors = [f'IbkrWsClient: Error message:  {message_data}']

        cm, success = self._send_payload(message_data, expected_errors=expected_errors)
        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def test_on_message_no_topic_handler(self):
        message_data = {'topic': 'unrecognized_topic', 'args': {'some_key': 'some_value'}}
        expected_errors = [f'IbkrWsClient: Topic "{message_data["topic"]}" unrecognised. Message: {message_data}']

        cm, success = self._send_payload(message_data, expected_errors=expected_errors)
        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def test_on_message_handled_without_subscription(self):
        message_data = {'topic': 'some_topic', 'args': {'channel': 'XYZ', 'data': 'info'}}
        expected_errors = [
            f'IbkrWsClient: Handled a channel "{message_data["topic"][1:]}" message that is missing a subscription. Message: {message_data}'
        ]

        with patch.object(self.ws_client, '_handle_subscribed_message', return_value=True):
            cm, success = self._send_payload(message_data, expected_errors=expected_errors)

        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def _logs_subscriptions(self, full_channel, data=None, needs_confirmation_sub: bool = False, needs_confirmation_unsub: bool = True):
        return [
            f'IbkrWsClient: Subscribed: s{full_channel}{"" if data is None else f"+{json.dumps(data)}"}{"" if not needs_confirmation_sub else " without confirmation."}',
            f'IbkrWsClient: Unsubscribed: u{full_channel}+{json.dumps(data if data is not None else {})}{"" if not needs_confirmation_unsub else " without confirmation."}',
        ]

    def test_on_message_market_data_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.MARKET_DATA)
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'data': {'fields': ['55', '71', '84', '86', '88', '85', '87', '7295', '7296', '70']}}
        response = {
            'topic': f's{full_channel}',
            'conid': self.conid,
            '_updated': self.update_time,
            55: 'AAPL',
            70: '195.34',
            71: '193.67',
            87: '24.2M',
            7295: '194.10',
            84: '195.25',
            86: '195.26',
            88: '3,500',
            85: '500',
            6508: '&serviceID1=122&serviceID2=123&serviceID3=203&serviceID4=775&serviceID5=204&serviceID6=206&serviceID7=108&serviceID8=109',
        }

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel, request['data']), [r.msg for r in cm.records])

        self.assertEqual(
            {
                self.conid: {
                    '_updated': self.update_time,
                    'conid': self.conid,
                    'topic': f'smd+{self.conid}',
                    'ask_price': '195.26',
                    'ask_size': '500',
                    'bid_price': '195.25',
                    'bid_size': '3,500',
                    'high': '195.34',
                    'low': '193.67',
                    'open': '194.10',
                    'service_params': '&serviceID1=122&serviceID2=123&serviceID3=203&serviceID4=775&serviceID5=204&serviceID6=206&serviceID7=108&serviceID8=109',
                    'symbol': 'AAPL',
                    'volume': '24.2M',
                }
            },
            queue.get(),
        )

    def test_on_message_market_history_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.MARKET_HISTORY)
        server_id = 87567
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {
            'channel': f'{full_channel}',
            'data': {'period': '1min', 'bar': '1min', 'outsideRTH': True, 'source': 'trades', 'format': '%o/%c/%h/%l'},
            'confirms_unsubscription': False,
        }
        response = {'topic': f's{full_channel}', 'serverId': server_id, '_updated': self.update_time, 'conid': self.conid, 'foo': 'bar'}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel, request['data']), [r.msg for r in cm.records])

        self.assertEqual(response, queue.get())
        self.assertIn(server_id, self.ws_client.server_ids(IbkrWsKey.MARKET_HISTORY))

    def test_on_message_trade_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.TRADES)
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}'}
        response = {'topic': f's{full_channel}', '_updated': self.update_time, 'conid': self.conid, 'args': [{'foo': 'bar'}]}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel), [r.msg for r in cm.records])
        self.assertEqual(response, queue.get())

    def test_on_message_orders_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.ORDERS)

        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}'}
        response = {'topic': f's{full_channel}', '_updated': self.update_time, 'conid': self.conid, 'args': [{'foo': 'bar'}]}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel, None, True, True), [r.msg for r in cm.records])
        self.assertEqual(response, queue.get())

    def test_subscription_without_confirmation(self):
        channel = 'fake'
        full_channel = f'{channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'needs_confirmation': False, 'confirms_unsubscription': False}
        response = None

        expected_errors = [f'IbkrWsClient: Channel subscription timeout: s{full_channel} after {self.subscription_retries} attempts.']

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response, expected_errors=expected_errors)
            self.assertTrue(success)

        self.assertEqual(
            [
                f'IbkrWsClient: Subscribed: s{full_channel} without confirmation.',
                f'IbkrWsClient: Unsubscribed: u{full_channel}+{{}} without confirmation.',
            ],
            [r.msg for r in cm.records],
        )

    def test_check_health(self):
        start_time = [100]
        has_active_connection_counter = [0]

        # control time
        def fake_time():
            start_time[0] += 100
            return start_time[0]

        # simulate that we don't have ws connection first
        def has_active_connection():
            has_active_connection_counter[0] += 1
            if has_active_connection_counter[0] <= 2:
                return False
            return True

        # prepare a fake subscription
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.TRADES)
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'data': {'foo': 'bar'}}
        response = {'topic': f's{full_channel}', '_updated': self.update_time, 'conid': self.conid, 'args': [{'foo': 'bar'}]}

        def run():
            # ensures each time WebSocketApp's mock is created, we override its on_message method
            def override_init_wsa_mock(wsa_mock: MagicMock, *args, **kwargs):
                wsa_mock = init_wsa_mock(wsa_mock, *args, **kwargs)
                wsa_mock._on_message.side_effect = lambda wsa_mock, message: wsa_mock.__on_message__(wsa_mock, json.dumps(response))
                return wsa_mock

            self.ws_client.start()
            self.ws_client.check_health()
            self.wsa_mock._on_message.side_effect = lambda wsa_mock, message: wsa_mock.__on_message__(wsa_mock, json.dumps(response))

            # create the original subscription
            self.ws_client.subscribe(**request)

            # we simulate that closing the WebSocket doesn't work since we have connectivity issues
            # self.wsa_mock.on_close.side_effect = lambda x, y, z: None

            # override time.time, ignore check_ping and take control of has_active_connection
            with patch('ibind.client.ibkr_ws_client.time') as time_mock, \
                    patch.object(self.ws_client, 'check_ping', return_value=True), \
                    patch('ibind.base.ws_client.WebSocketApp', side_effect=lambda *args, **kwargs: override_init_wsa_mock(self.wsa_mock, *args, **kwargs)), \
                    patch.object(self.ws_client, '_has_active_connection', side_effect=has_active_connection) as has_active_connection_mock:  # fmt: skip
                time_mock.time.side_effect = fake_time
                self.ws_client._last_heartbeat = self.max_ping_interval * 1000

                # this should try to close the connection, fail to do so, abandon the WebSocketApp's mock,
                # then recreate a new mock and recreate the connections
                self.ws_client.check_health()

                self.assertTrue(self.ws_client.ready())
                self.assertEqual([call()] * 6, has_active_connection_mock.call_args_list)
                self.ws_client.shutdown()

        expected_errors = [
            f'IbkrWsClient: Last IBKR heartbeat happened 162.00 seconds ago, exceeding the max ping interval of {self.max_ping_interval}. Restarting.',
            # 'IbkrWsClient: Hard reset close timeout',
            # f'IbkrWsClient: Abandoning current WebSocketApp that cannot be closed: {self.wsa_mock}'
        ]

        cm, success = self.run_in_test_context(run, expected_errors=expected_errors)

        channel_subscribed_log = f'IbkrWsClient: Subscribed: s{full_channel}+{json.dumps(request["data"])}'

        self.assertEqual(
            [channel_subscribed_log]
            + expected_errors
            + [
                f'IbkrWsClient: Invalidated subscription: {full_channel}',
                f"IbkrWsClient: Recreating 1/1 subscriptions: {{'{full_channel}': {{'status': False, 'data': {request['data']}, 'needs_confirmation': True, 'subscription_processor': None}}}}",
                channel_subscribed_log,
                f'IbkrWsClient: Invalidated subscription: {full_channel}',
            ],
            [r.msg for r in cm.records],
        )



from pprint import pformat
from unittest import TestCase
from unittest.mock import MagicMock, patch, call

from ibind.base.rest_client import Result
from ibind.client.ibkr_utils import StockQuery, filter_stocks, find_answer, QuestionType, handle_questions, question_type_to_message_id, OrderRequest, parse_order_request
from ibind.support.logs import project_logger
from test.integration.client import ibkr_responses
from test_utils import verify_log


class TestIbkrUtilsI(TestCase):
    def setUp(self):
        self.instruments = ibkr_responses.responses['stocks']
        self.result = Result(data=self.instruments)
        self.maxDiff = None

    def test_filter_stocks(self):
        queries = [
            StockQuery(symbol='AAPL', contract_conditions={'isUS': False}, name_match='APPLE'),
            StockQuery(symbol='BBVA', contract_conditions={'exchange': 'NYSE'}),
            StockQuery(symbol='CDN', contract_conditions={'isUS': True}),
            StockQuery(symbol='CFC', contract_conditions={}),
            StockQuery(symbol='GOOG', contract_conditions={'isUS': False}, instrument_conditions={'chineseName': 'Alphabet&#x516C;&#x53F8;'}),
            'HUBS',
            StockQuery(symbol='META', name_match='meta ', contract_conditions={'isUS': False}, instrument_conditions={}),
            StockQuery(symbol='MSFT', contract_conditions={'exchange': 'NASDAQ'}),
            StockQuery(symbol='SAN', name_match='SANTANDER'),
            StockQuery(symbol='SCHW', contract_conditions={'exchange': 'NASDAQ'}),
            StockQuery(symbol='TEAM', name_match='ATLASSIAN'),
            StockQuery(symbol='INVALID_SYMBOL')
        ]  # fmt: skip
        with self.assertLogs(project_logger(), level='INFO') as cm:
            rv = filter_stocks(queries, Result(data=self.instruments), default_filtering=False)

        verify_log(
            self, cm, [f'Error getting stocks. Could not find valid instruments INVALID_SYMBOL in result: {self.result}. Skipping query={queries[-1]}.']
        )  # fmt: skip

        # pprint(rv)

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x82F9;&#x679C;&#x516C;&#x53F8;',
                    'contracts': [
                        {'conid': 38708077, 'exchange': 'MEXI', 'isUS': False},
                        {'conid': 273982664, 'exchange': 'EBS', 'isUS': False},
                    ],
                    'name': 'APPLE INC',
                },
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x82F9;&#x679C;&#x516C;&#x53F8;',
                    'contracts': [{'conid': 532640894, 'exchange': 'AEQLIT', 'isUS': False}],
                    'name': 'APPLE INC-CDR',
                },
            ],
            rv.data['AAPL'],
        )

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x897F;&#x73ED;&#x7259;&#x5BF9;&#x5916;&#x94F6;&#x884C;',
                    'contracts': [{'conid': 4815, 'exchange': 'NYSE', 'isUS': True}],
                    'name': 'BANCO BILBAO VIZCAYA-SP ADR',
                },
            ],
            rv.data['BBVA'],
        )

        self.assertEqual([], rv.data['CDN'])

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': None,
                    'contracts': [{'conid': 42001300, 'exchange': 'IBIS', 'isUS': False}],
                    'name': 'UET UNITED ELECTRONIC TECHNO',
                }
            ],
            rv.data['CFC'],
        )

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': 'Alphabet&#x516C;&#x53F8;',
                    'contracts': [
                        {'conid': 210810667, 'exchange': 'MEXI', 'isUS': False},
                    ],
                    'name': 'ALPHABET INC-CL C',
                },
                {
                    'assetClass': 'STK',
                    'chineseName': 'Alphabet&#x516C;&#x53F8;',
                    'contracts': [{'conid': 532638805, 'exchange': 'AEQLIT', 'isUS': False}],
                    'name': 'ALPHABET INC - CDR',
                },
            ],
            rv.data['GOOG'],
        )

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': 'HubSpot&#x516C;&#x53F8;',
                    'contracts': [{'conid': 169544810, 'exchange': 'NYSE', 'isUS': True}],
                    'name': 'HUBSPOT INC',
                }
            ],
            rv.data['HUBS'],
        )

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': 'Meta&#x5E73;&#x53F0;&#x80A1;&#x4EFD;&#x6709;&#x9650;&#x516C;&#x53F8;',
                    'contracts': [
                        {'conid': 114922621, 'exchange': 'MEXI', 'isUS': False},
                    ],
                    'name': 'META PLATFORMS INC-CLASS A',
                },
                {
                    'assetClass': 'STK',
                    'chineseName': 'Meta&#x5E73;&#x53F0;&#x80A1;&#x4EFD;&#x6709;&#x9650;&#x516C;&#x53F8;',
                    'contracts': [{'conid': 530091499, 'exchange': 'AEQLIT', 'isUS': False}],
                    'name': 'META PLATFORMS INC-CDR',
                },
            ],
            rv.data['META'],
        )

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x5FAE;&#x8F6F;&#x516C;&#x53F8;',
                    'contracts': [
                        {'conid': 272093, 'exchange': 'NASDAQ', 'isUS': True},
                    ],
                    'name': 'MICROSOFT CORP',
                },
            ],
            rv.data['MSFT'],
        )

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x6851;&#x5766;&#x5FB7;',
                    'contracts': [
                        {'conid': 38708867, 'exchange': 'MEXI', 'isUS': False},
                        {'conid': 385055564, 'exchange': 'WSE', 'isUS': False},
                    ],
                    'name': 'BANCO SANTANDER SA',
                },
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x6851;&#x5766;&#x5FB7;',
                    'contracts': [{'conid': 12442, 'exchange': 'NYSE', 'isUS': True}],
                    'name': 'BANCO SANTANDER SA-SPON ADR',
                },
                {
                    'assetClass': 'STK',
                    'chineseName': '&#x6851;&#x5766;&#x5FB7;&#x82F1;&#x56FD;&#x516C;&#x5171;&#x6709;&#x9650;&#x516C;&#x53F8;',
                    'contracts': [{'conid': 80993135, 'exchange': 'LSE', 'isUS': False}],
                    'name': 'SANTANDER UK PLC',
                },
            ],
            rv.data['SAN'],
        )

        self.assertEqual([], rv.data['SCHW'])

        self.assertEqual(
            [
                {
                    'assetClass': 'STK',
                    'chineseName': None,
                    'contracts': [{'conid': 589316251, 'exchange': 'NASDAQ', 'isUS': True}],
                    'name': 'ATLASSIAN CORP-CL A',
                },
            ],
            rv.data['TEAM'],
        )

    def test_question_type_to_message_id_successful(self):
        question_type = QuestionType.PRICE_PERCENTAGE_CONSTRAINT
        message_id = question_type_to_message_id(question_type)
        self.assertEqual(message_id, 'o163')


class TestFindAnswer(TestCase):
    def setUp(self):
        # Setup Answers dictionary here
        self.answers = {QuestionType.PRICE_PERCENTAGE_CONSTRAINT: True}

    def test_valid_question(self):
        question = f'Some {QuestionType.PRICE_PERCENTAGE_CONSTRAINT} specific question'
        answer = find_answer(question, self.answers)
        self.assertTrue(answer)

    def test_invalid_question(self):
        question = 'Nonexistent question type'
        with self.assertRaises(ValueError):
            find_answer(question, self.answers)


class TestHandleQuestionsI(TestCase):
    def setUp(self):
        self.original_result = Result(
            data=[{'id': '12345', 'message': ['price exceeds the Percentage constraint of 3%.']}], request={'url': 'test_url'}
        )
        self.answers = {QuestionType.PRICE_PERCENTAGE_CONSTRAINT: True}
        self.reply_callback = MagicMock()

    @patch('ibind.client.ibkr_utils.QuestionType')
    def test_successful_handling(self, question_type_mock):
        # Mocking the QuestionType enum
        question_type_mock.PRICE_PERCENTAGE_CONSTRAINT.__str__.return_value = 'price exceeds the Percentage constraint of 3%.'
        question_type_mock.ADDITIONAL_QUESTION_TYPE.__str__.return_value = 'This is an additional question.'

        self.answers = {question_type_mock.PRICE_PERCENTAGE_CONSTRAINT: True, question_type_mock.ADDITIONAL_QUESTION_TYPE: True}

        # Mock reply_callback to simulate the sequence of question-answer interactions
        replies = [
            Result(data=[{'id': '12346', 'message': ['This is an additional question.']}], request={'url': 'another_question_url'}),
            Result(data=[{'id': '12347'}], request={'url': 'final_url'}),  # No more questions
        ]
        self.reply_callback.side_effect = replies

        result = handle_questions(self.original_result, self.answers, self.reply_callback)
        self.assertEqual(result.request['url'], self.original_result.request['url'])
        self.assertEqual(len(self.reply_callback.call_args_list), 2)
        # Expected calls to self.reply_callback
        expected_calls = [
            call(
                self.original_result.data[0]['id'], self.answers[question_type_mock.PRICE_PERCENTAGE_CONSTRAINT]
            ),  # First call with question ID '12346' and reply True
            call(
                replies[0].data[0]['id'], self.answers[question_type_mock.ADDITIONAL_QUESTION_TYPE]
            ),  # Second call with question ID '12347' and reply True
        ]

        # Check if the calls to self.reply_callback are as expected
        self.assertEqual(expected_calls, self.reply_callback.call_args_list)

    def test_too_many_questions(self):
        # Simulate repetitive questions to exceed the question limit
        self.reply_callback.side_effect = [self.original_result] * 21

        with self.assertRaises(RuntimeError) as cm_err:
            handle_questions(self.original_result, self.answers, self.reply_callback)

        self.assertIn('Too many questions', str(cm_err.exception))

    def test_negative_reply(self):
        # Set a negative answer
        self.answers[QuestionType.PRICE_PERCENTAGE_CONSTRAINT] = False

        with self.assertRaises(RuntimeError) as cm_err:
            handle_questions(self.original_result, self.answers, self.reply_callback)
        self.assertEqual(
            f'A question was not given a positive reply. Question: "{self.original_result.data[0]["message"][0]}". Answers: \n{self.answers}\n. Request: {self.original_result.request}',
            str(cm_err.exception),
        )

    def test_multiple_orders_returned(self):
        # Simulate multiple orders in the data
        self.original_result.data = [
            {'id': '12345', 'message': [str(QuestionType.PRICE_PERCENTAGE_CONSTRAINT)]},
            {'id': '12346', 'message': [str(QuestionType.PRICE_PERCENTAGE_CONSTRAINT)]},
        ]
        self.reply_callback.return_value = self.original_result.copy(data=[{}])

        with self.assertLogs(project_logger(), level='INFO') as cm:
            handle_questions(self.original_result, self.answers, self.reply_callback)

        verify_log(self, cm, ['While handling questions multiple orders were returned: ' + pformat(self.original_result.data)])

    def test_multiple_messages_returned(self):
        # Simulate a single order with multiple messages
        self.original_result.data = [{'id': '12345', 'message': [str(QuestionType.PRICE_PERCENTAGE_CONSTRAINT), 'Message 2']}]
        self.reply_callback.return_value = self.original_result.copy(data=[{}])

        with self.assertLogs(project_logger(), level='INFO') as cm:
            handle_questions(self.original_result, self.answers, self.reply_callback)

        verify_log(self, cm, ['While handling questions multiple messages were returned: ' + pformat(self.original_result.data[0]['message'])])

class TestParseOrderRequestI(TestCase):
    def test_parse_both_with_conidex(self):
        order_request = OrderRequest(
            conid=None,
            side='BUY',
            quantity=321,
            order_type='MKT',
            acct_id='DU1234567',
            conidex='33333' # should cause exception
        )

        d = parse_order_request(order_request)

        self.assertEqual({
            'side': 'BUY',
            'quantity': 321,
            'orderType': 'MKT',
            'acctId': 'DU1234567',
            'conidex': '33333',
            'tif': 'GTC'
        }, d)

    def test_raise_with_conid_and_conidex(self):
        with self.assertRaises(ValueError) as cm_err:
            order_request = OrderRequest(
                conid=123,
                side='BUY',
                quantity=321,
                order_type='MKT',
                acct_id='DU1234567',
                conidex='33333' # should cause exception
            )

            parse_order_request(order_request)

        self.assertEqual("Both 'conidex' and 'conid' are provided. When using 'conidex', specify `conid=None`.", str(cm_err.exception))





import functools
import logging
import sys
import traceback
import types
import unittest
from unittest import TestCase
from unittest._log import _CapturingHandler, _AssertLogsContext

from ibind.support.py_utils import make_clean_stack


def raise_from_context(cm, level='WARNING'):
    for record in cm.records:
        if record.levelno >= getattr(logging, level):
            raise RuntimeError(record.message)


def verify_log(test_case: TestCase, cm, expected_messages, comparison: callable = lambda x, y: x == y):
    messages = [record.msg for record in cm.records]
    missing_expected = expected_messages.copy()
    for i, expected_msg in enumerate(expected_messages):
        for msg in messages:
            if comparison(expected_msg, msg):
                missing_expected.remove(expected_msg)
                break

    if missing_expected:
        test_case.fail('Expected log(s) not found:\n\t{}'.format('\n\t'.join(missing_expected)))


def verify_log_simple(test_self, cm, expected_messages):
    for i, msg in enumerate(expected_messages):
        test_self.assertEqual(msg, cm.records[i].msg)


def exact_log(test_case, cm, expected_messages):
    test_case.assertEqual(expected_messages, [record.msg for record in cm.records])


class SafeAssertLogs(_AssertLogsContext):
    """
    The self.assertLogs context manager, that sets log level on the handler instead of logger.

    Original docstring:
    A context manager used to implement TestCase.assertLogs().
    """

    def __init__(self, *args, logger_level: str = None, **kwargs):
        if sys.version_info < (3, 10, 0) and 'no_logs' in kwargs:
            del kwargs['no_logs']

        super().__init__(*args, **kwargs)
        self.logger_level = logger_level

    def __enter__(self, include_original_handlers: bool = False):
        if isinstance(self.logger_name, logging.Logger):
            logger = self.logger = self.logger_name
        else:
            logger = self.logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(self.LOGGING_FORMAT)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        self.watcher = handler.watcher
        self.old_handlers = logger.handlers[:]
        self.old_level = logger.level
        self.old_propagate = logger.propagate
        logger.handlers = [handler]
        handler.setLevel(self.level)  # this one line is different, originally was `logger.setLevel`
        logger.propagate = False
        if self.logger_level is not None:
            logger.setLevel(getattr(logging, self.logger_level))

        if include_original_handlers:
            logger.handlers += self.old_handlers
            logger.propagate = True
        return handler.watcher


def get_logger_children(main_logger) -> list[logging.Logger]:
    """
    Gets child loggers. Added as a support compat for Python version 3.11 and below.
    Source: https://github.com/python/cpython/blob/3.12/Lib/logging/__init__.py#L1831
    """

    def _hierlevel(logger):
        if logger is logger.manager.root:
            return 0
        return 1 + logger.name.count('.')

    d = main_logger.manager.loggerDict
    # exclude PlaceHolders - the last check is to ensure that lower-level
    # descendants aren't returned - if there are placeholders, a logger's
    # parent field might point to a grandparent or ancestor thereof.
    return [
        item
        for item in d.values()
        if isinstance(item, logging.Logger) and item.parent is main_logger and _hierlevel(item) == 1 + _hierlevel(item.parent)
    ]


class RaiseLogsContext:
    """
    Captures log messages at or above a specified level and raises unexpected ones as exceptions.

    This context manager monitors log messages from a specified logger. Any log messages
    at or above the given logging level are recorded. If a message is not explicitly
    expected, a `RuntimeError` is raised, including the stack trace of the log call. It ensures
    loggers are restored to their original state after use.

    Note:
        - When used in conjunction with `self.assertLogs` or `SafeAssertLogs`, ensure this context manager is defined last to properly assert log expectations.

    Args:
        test_case (TestCase): The test case instance, typically from `unittest.TestCase`.
        logger_name (str | None): The name of the logger to monitor. Defaults to the root logger.
        level (str): The logging level threshold (e.g., 'ERROR', 'WARNING'). Logs at or above this level are captured.
        expected_errors (list[str] | None): A list of log messages that are expected and should not trigger an exception.
        comparison (Callable[[str, str], bool]): A function to compare expected errors with log messages.
            Defaults to an exact string match (`lambda x, y: x == y`).

    Example Usage:
        >>> with RaiseLogsContext(self, logger_name='my_logger', level='WARNING', expected_errors=['My expected warning']):
        ...     logging.getLogger('my_logger').warning('My expected warning')  # No error
        ...     logging.getLogger('my_logger').error('Unexpected issue')  # Raises RuntimeError
    """

    def __init__(
        self,
        test_case: TestCase,
        logger_name=None,
        level='ERROR',
        expected_errors: [str] = None,
        comparison: callable = lambda x, y: x == y,
    ):
        self._test_case = test_case
        self._logger_name = logger_name
        self._level = level
        self._level_no = getattr(logging, level)
        if expected_errors is None:
            expected_errors = []
        self._expected_errors = expected_errors
        self._comparison = comparison

    def monkey_patch_log(self, original_method):
        """Wraps a logger method to attach a manually captured stack trace to log records."""

        def new_method(msg, *args, **kwargs):
            # Store the manually captured stack trace in the log record
            stack = make_clean_stack()
            if 'extra' not in kwargs:
                kwargs['extra'] = {}
            kwargs['extra']['manual_trace'] = stack

            # Call the original logging method with the modified arguments
            return original_method(msg, *args, **kwargs)

        return new_method

    def monkey_patch_loggers(self, loggers):
        """Monkey-patches loggers to attach a stack trace to warning and error messages."""
        for logger in loggers:
            if self._level_no <= logging.ERROR:
                logger.__old_error_method__ = logger.error
                logger.error = self.monkey_patch_log(logger.error)

            if self._level_no <= logging.WARNING:
                logger.__old_warning_method__ = logger.warning
                logger.warning = self.monkey_patch_log(logger.warning)

    def restore_loggers(self, loggers):
        """Restores the original error and warning logging methods after patching."""
        for logger in loggers:
            if self._level_no <= logging.ERROR:
                logger.error = logger.__old_error_method__  # Restore the original error method

            if self._level_no <= logging.WARNING:
                logger.warning = logger.__old_warning_method__  # Restore the original warning method

    def __enter__(self):
        """
        Initializes the logging context by patching loggers and setting up a log watcher.

        This method ensures that logs at the specified level are captured and asserts
        that unexpected log messages are raised as errors.
        """

        self._logger = logging.getLogger(self._logger_name)
        loggers_to_be_patched = [self._logger] + get_logger_children(self._logger)
        self.monkey_patch_loggers(loggers_to_be_patched)  # Apply monkey-patching to attach stack traces to logged messages

        # Initialize SafeAssertLogs, a helper to capture and assert log records
        self._context_manager = SafeAssertLogs(self._test_case, self._logger, level=self._level, no_logs=False)

        # Enter the SafeAssertLogs context, starting log capture and returning the watcher
        self._watcher = self._context_manager.__enter__(include_original_handlers=True)
        return self._watcher

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Restores original logger methods and verifies captured log messages.

        This method is called when exiting the context manager. It ensures that:
        - Monkey-patched loggers are restored to their original state.
        - If an exception occurred inside the `with` block, it is propagated normally.
        - If no exception occurred, all captured log messages are checked against expected errors.
        - Unexpected log messages result in a `RuntimeError`.
        """

        # Restore original logging methods that were monkey-patched
        loggers_to_be_patched = [self._logger] + get_logger_children(self._logger)
        self.restore_loggers(loggers_to_be_patched)

        # If an exception occurred inside the 'with' block, return False to let Python re-raise it
        if exc_type is not None:
            return False

        # If no logs were captured return True to indicate that no errors were encountered and that the context exited cleanly
        if len(self._watcher.records) == 0:
            return True

        for record in self._watcher.records:
            found = False

            # Check if the log message matches any of the expected error messages
            for expected_error in self._expected_errors:
                if self._comparison(expected_error, record.msg):
                    found = True
                    break

            # If the message is expected, move on to the next record
            if found:
                continue

            # If the log record has a manually stored traceback, raise an error with that traceback
            if hasattr(record, 'manual_trace'):
                raise RuntimeError(
                    '\n' + ''.join(traceback.format_list(record.manual_trace)) + f'Logger {self._logger} logged an unexpected message:\n{record.msg}'
                )

            # Otherwise, raise an error using the log record's location
            raise RuntimeError(f'\n...\nFile "{record.pathname}", line {record.lineno} in {record.funcName}\n{record.msg}')


def raise_logs(level='ERROR', logger_name=None):
    def _wrapper(fn):
        @functools.wraps(fn)
        def wrapper(self, *args, **kwargs):
            with RaiseLogsContext(self, level=level, logger_name=logger_name):
                return fn(self, *args, **kwargs)

        return wrapper

    return _wrapper


def decorate_methods(decorator, starts_with=''):
    class DecorateMethods(type):
        """Decorate all methods of the class with the decorator provided"""

        def __new__(cls, name, bases, attrs, **kwargs):
            exclude = kwargs.get('exclude', [])

            for attr_name, attr_value in attrs.items():
                if (
                    isinstance(attr_value, types.FunctionType)
                    and attr_name.startswith(starts_with)
                    and attr_name not in exclude
                    and not hasattr(attr_value, '__exclude_decorator__')
                    and not attr_name.startswith('__')
                ):
                    attrs[attr_name] = decorator(attr_value)

            return super(DecorateMethods, cls).__new__(cls, name, bases, attrs)

    return DecorateMethods


class TestCaseWithRaiseLogs(unittest.TestCase, metaclass=decorate_methods(raise_logs(logger_name='ibind'), starts_with='test')): ...


def exclude_decorator(fn):
    fn.__exclude_decorator__ = True
    return fn



import datetime
import pprint
import threading
from dataclasses import dataclass, field, fields
from typing import Optional, Dict, Union, TYPE_CHECKING
from warnings import warn

from ibind.base.rest_client import Result, pass_result
from ibind.client.ibkr_definitions import decode_data_availability
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger
from ibind.support.py_utils import UNDEFINED, ensure_list_arg, VerboseEnum, OneOrMany, exception_to_string
from ibind import var

_LOGGER = project_logger(__file__)

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient


@dataclass
class StockQuery:
    """
    A class to encapsulate query parameters for filtering stock data.

    This class is used to define a set of criteria for filtering stocks, which includes the stock symbol,
    name matching pattern, and conditions for instruments and contracts.

    Attributes:
        symbol (str): The stock symbol to query.
        name_match (Optional[str], optional): A string pattern to match against stock names. Optional.
        instrument_conditions (Optional[dict], optional): Key-value pairs representing conditions to apply to
            stock instruments. Each condition is matched exactly against the instrument's attributes.
        contract_conditions (Optional[dict], optional): Key-value pairs representing conditions to apply to
            stock contracts. Each condition is matched exactly against the contract's attributes.
    """

    symbol: str
    name_match: Optional[str] = field(default=None)
    instrument_conditions: Optional[dict] = field(default=None)
    contract_conditions: Optional[dict] = field(default=None)


StockQueries = OneOrMany[Union[StockQuery, str]]


def _filter(data: dict, conditions: dict) -> bool:
    for key, value in conditions.items():
        if data.get(key, UNDEFINED) != value:
            return False

    return True


def process_instruments(
    instruments: dict,
    name_match: str = None,
    instrument_conditions: dict = None,
    contract_conditions: dict = None,
) -> [dict]:
    """
    Filters a list of instruments based on specified name matching and conditions.

    This function processes each instrument in the given list and filters them based on the provided
    name matching pattern, instrument conditions, and contract conditions. Instruments are filtered
    to match the specified conditions exactly.

    Parameters:
        instruments (List[dict]): A list of instrument dictionaries to be filtered.
        name_match (Optional[str], optional): A string pattern for partial matching of instrument names. If specified, only instruments  with names containing this pattern are included.
        instrument_conditions (Optional[dict], optional): Key-value pairs representing exact conditions to match against each instrument's attributes.
        contract_conditions (Optional[dict], optional): Key-value pairs representing exact conditions to match against each contract within an instrument.

    Returns:
        List[dict]: A filtered list of instruments that meet the specified criteria.
    """
    filtered_instruments = []

    for instrument in instruments:
        # look for a partial match of the instrument name if provided
        if name_match is not None and name_match.upper() not in instrument['name'].upper():
            continue

        # look for an exact match of instrument properties if provided
        if instrument_conditions is not None and not _filter(instrument, instrument_conditions):
            continue

        # filter contracts by conditions provided
        if contract_conditions is not None:
            filtered_contracts = list(
                filter(
                    lambda x: _filter(x, contract_conditions),
                    instrument['contracts'],
                )
            )

            # if no contracts are left, we don't need the instrument
            if not filtered_contracts:
                continue

            # if all conditions are  met, accept the instrument and its contracts
            instrument['contracts'] = filtered_contracts

        filtered_instruments.append(instrument)

    return filtered_instruments


@ensure_list_arg('queries')
def filter_stocks(queries: StockQueries, result: Result, default_filtering: bool = True):
    stocks = {}
    data = result.data
    for q in queries:
        symbol, name_match, instrument_conditions, contract_conditions = process_query(q, default_filtering)

        if symbol not in data or len(data[symbol]) == 0:
            _LOGGER.error(f'Error getting stocks. Could not find valid instruments {symbol} in result: {result}. Skipping query={q}.')
            continue

        filtered_instruments = process_instruments(data[symbol], name_match, instrument_conditions, contract_conditions)

        stocks[symbol] = filtered_instruments

    return pass_result(stocks, result)


def query_to_symbols(queries):
    return ','.join([q if isinstance(q, str) else q.symbol for q in queries])


def process_query(q, default_filtering: bool = True):
    if isinstance(q, str):
        q = StockQuery(symbol=q)

    symbol = q.symbol
    name_match = q.name_match
    instrument_conditions = q.instrument_conditions
    # look for US contracts only by default
    default_contract_filter = {'isUS': True} if default_filtering else None
    contract_conditions = q.contract_conditions
    if contract_conditions is None:
        contract_conditions = default_contract_filter

    return symbol, name_match, instrument_conditions, contract_conditions


class QuestionType(VerboseEnum):
    """
    Enumeration of common warning messages encountered when submitting orders.

    This enum class represents different types of precautionary messages that may be returned by IBKR's API when placing an order. These warnings often require user confirmation before proceeding.

    Note:
        - Look for all suppressible message IDs in IBKR API documentation (https://www.interactivebrokers.com/campus/ibkr-api-page/cpapi-v1/#suppressible-id)
    """

    PRICE_PERCENTAGE_CONSTRAINT = 'price exceeds the Percentage constraint of 3%'
    MISSING_MARKET_DATA = (
        'You are submitting an order without market data. We strongly recommend against this as it may result in erroneous and unexpected trades.'
    )
    TICK_SIZE_LIMIT = 'exceeds the Tick Size Limit of'
    ORDER_SIZE_LIMIT = 'size exceeds the Size Limit of'
    TRIGGER_AND_FILL = 'This order will most likely trigger and fill immediately.'
    ORDER_VALUE_LIMIT = 'exceeds the Total Value Limit of'
    SIZE_MODIFICATION_LIMIT = 'size modification exceeds the size modification limit'
    MANDATORY_CAP_PRICE = 'To avoid trading at a price that is not consistent with a fair and orderly market'
    CASH_QUANTITY = 'Traders are responsible for understanding cash quantity details, which are provided on a best efforts basis only.'
    CASH_QUANTITY_ORDER = 'Orders that express size using a monetary value (cash quantity) are provided on a non-guaranteed basis.'
    STOP_ORDER_RISKS = (
        'You are about to submit a stop order. Please be aware of the various stop order types available and the risks associated with each one.'
    )
    MULTIPLE_ACCOUNTS = 'This order will be distributed over multiple accounts. We strongly suggest you familiarize yourself with our allocation facilities before submitting orders.'
    DISRUPTIVE_ORDERS = 'If your order is not immediately executable, our systems may, depending on market conditions, reject your order'
    CLOSE_POSITION = 'Would you like to cancel all open orders and then place new closing order?'


# FIXME: Fill in the remaining question types as we find out what they are
_MESSAGE_ID_TO_QUESTION_TYPE = {
    "o163": (QuestionType.PRICE_PERCENTAGE_CONSTRAINT, "The following order exceeds the price percentage limit"),
    "o354": (QuestionType.MISSING_MARKET_DATA, "You are submitting an order without market data. We strongly recommend against this as it may result in erroneous and unexpected trades. Are you sure you want to submit this order?"),
    "o382": (QuestionType.TICK_SIZE_LIMIT, "The following value exceeds the tick size limit"),
    "o383": (QuestionType.ORDER_SIZE_LIMIT, "The following order BUY 650 AAPL NASDAQ.NMS size exceeds the Size Limit of 500.\nAre you sure you want to submit this order?"),
    "o403": (QuestionType.TRIGGER_AND_FILL, "This order will most likely trigger and fill immediately.\nAre you sure you want to submit this order?"),
    "o451": (QuestionType.ORDER_VALUE_LIMIT, "The following order BUY 650 AAPL NASDAQ.NMS value estimate of 124,995.00 USD exceeds \nthe Total Value Limit of 100,000 USD.\nAre you sure you want to submit this order?"),
    "o2136": (UNDEFINED, "Mixed allocation order warning"),
    "o2137": (UNDEFINED, "Cross side order warning"),
    "o2165": (UNDEFINED, "Warns that instrument does not support trading in fractions outside regular trading hours"),
    "o10082": (UNDEFINED, "Called Bond warning"),
    "o10138": (QuestionType.SIZE_MODIFICATION_LIMIT, "The following order size modification exceeds the size modification limit."),
    "o10151": (UNDEFINED, "Warns about risks with Market Orders"),
    "o10152": (UNDEFINED, "Warns about risks associated with stop orders once they become active"),
    "o10153": (QuestionType.MANDATORY_CAP_PRICE, "<h4>Confirm Mandatory Cap Price</h4>To avoid trading at a price that is not consistent with a fair and orderly market, IB may set a cap (for a buy order) or sell order). THIS MAY CAUSE AN ORDER THAT WOULD OTHERWISE BE MARKETABLE TO NOT BE TRADED."),
    "o10164": (QuestionType.CASH_QUANTITY, "Traders are responsible for understanding cash quantity details, which are provided on a best efforts basis only."),
    "o10223": (QuestionType.CASH_QUANTITY_ORDER, "<h4>Cash Quantity Order Confirmation</h4>Orders that express size using a monetary value (cash quantity) are provided on a non-guaranteed basis. The system simulates the order by cancelling it once the specified amount is spent (for buy orders) or collected (for sell orders). In addition to the monetary value, the order uses a maximum size that is calculated using the Cash Quantity Estimate Factor, which you can modify in Presets."),
    "o10288": (UNDEFINED, "Warns about risks associated with market orders for Crypto"),
    "o10331": (QuestionType.STOP_ORDER_RISKS, "You are about to submit a stop order. Please be aware of the various stop order types available and the risks associated with each one.\nAre you sure you want to submit this order?"),
    "o10332": (UNDEFINED, "OSL Digital Securities LTD Crypto Order Warning"),
    "o10333": (UNDEFINED, "Option Exercise at the Money warning"),
    "o10334": (UNDEFINED, "Warns that order will be placed into current omnibus account instead of currently selected global account."),
    "o10335": (UNDEFINED, "Serves internal Rapid Entry window."),
    "p6": (QuestionType.MULTIPLE_ACCOUNTS, "This order will be distributed over multiple accounts. We strongly suggest you familiarize yourself with our allocation facilities before submitting orders."),
    "p12": (QuestionType.DISRUPTIVE_ORDERS, "If your order is not immediately executable, our systems may, depending on market conditions, reject your order if its limit price is more than the allowed amount away from the reference price at that time. If this happens, you will not receive a fill. This is a control designed to ensure that we comply with our regulatory obligations to avoid submitting disruptive orders to the marketplace.\\nUse the Price Management Algo?"),
}

_QUESTION_TYPE_TO_MESSAGE_ID = {v[0]: k if v[0] is not UNDEFINED else 'undefined' for k, v in _MESSAGE_ID_TO_QUESTION_TYPE.items()}


def question_type_to_message_id(question_type: QuestionType) -> str:
    """
    Converts a QuestionType to its corresponding message ID.

    Parameters:
        question_type (QuestionType): The QuestionType enum value.

    Returns:
        str: The corresponding message ID string.

    Note:
        - Look for all suppressible message IDs in IBKR API documentation (https://www.interactivebrokers.com/campus/ibkr-api-page/cpapi-v1/#suppressible-id)
    """
    if question_type not in _QUESTION_TYPE_TO_MESSAGE_ID:
        raise ValueError(f'QuestionType {question_type} is not currently dynamically mapped to a message id. Please look the ID up manually.')

    return _QUESTION_TYPE_TO_MESSAGE_ID[question_type]


Answers = Dict[Union[QuestionType, str], bool]
"""
A mapping of order warnings to user responses.

This dictionary type is used to associate specific warning messages from the IBKR API (either predefined in `QuestionType` or as raw strings) with a boolean response
indicating whether the user accepts or rejects the warning.

Key:
    - `QuestionType`: A predefined enum representing common IBKR warning messages.
    - `str`: A raw string warning message (if not covered by `QuestionType`).

Value:
    - `bool`:
        - `True` if the user acknowledges and accepts the warning.
        - `False` if the user rejects it, potentially preventing the order submission.

Example:
    >>> user_answers = {
    ...     QuestionType.PRICE_PERCENTAGE_CONSTRAINT: True,
    ...     "Some custom warning message": False
    ... }
"""


def find_answer(question: str, answers: Answers):
    """
    Retrieves a predefined answer for a given question based on question types defined in the
    QuestionType enum.

    This function matches the given question against known question types and returns the corresponding
    predefined answer. It uses the 'QuestionType' enum to identify types of questions and their
    associated answers.

    Parameters:
        question (str): The question for which an answer is sought.
        answers (Answers): A dictionary mapping QuestionType enum members to their corresponding boolean answers.

    Returns:
        bool: The predefined answer (boolean) corresponding to the given question.

    Raises:
        ValueError: If no predefined answer is found for the given question.
    """
    for question_type, answer in answers.items():
        if str(question_type) in question:
            return answer

    raise ValueError(f'No answer found for question: "{question}"')


def handle_questions(original_result: Result, answers: Answers, reply_callback: callable) -> Result:
    """
    Handles a series of interactive questions that may arise during a request, especially when submitting orders.

    This method iteratively processes questions contained within the response data. It expects
    each question to be answered affirmatively to proceed. If a question does not receive
    a positive reply or if there are too many questions (more than 10 attempts), a RuntimeError
    is raised.

    Parameters:
        original_result (support.rest_client.Result): The initial result object containing the data that may include questions.
        answers (Answers): A collection of answers to the expected questions.

    Returns:
        support.rest_client.Result: The updated result object after all questions have been successfully answered.

    Raises:
        RuntimeError: If a question does not receive a positive reply or if there are too many questions.

    See:
        QuestionType: for types of answers currently supported.

    Note:
        - The function assumes that each response will contain at most one question.
    """

    result: Result = original_result.copy()

    questions = []  # we store questions in case we need to show them at the end
    for attempt in range(20):
        data = result.data

        if 'error' in data:
            if "Order couldn't be submitted: Local order ID=" in data['error']:
                raise ExternalBrokerError(f"Order couldn't be submitted. Orders are already registered: {original_result.request.get('json', {}).get('orders', {})}")

            raise ExternalBrokerError(f'While handling questions an error was returned: {pprint.pformat(data)}')

        if not isinstance(data, list):
            raise ExternalBrokerError(f'While handling questions unknown data was returned: {data!r}. Request: {result.request}')

        first_data = data[0]  # this assumes submitting only one order at a time

        # we interpret messages as questions, absence of which we interpret as the end of questions
        if 'message' not in first_data:
            if len(data) == 1:
                data = data[0]
            return pass_result(data, original_result)

        if len(data) != 1:
            _LOGGER.warning(f'While handling questions multiple orders were returned: {pprint.pformat(data)}')

        messages = first_data['message']

        if len(messages) != 1:
            _LOGGER.warning(f'While handling questions multiple messages were returned: {pprint.pformat(messages)}')

        question = messages[0]
        question = question.strip().replace('\n', '')  # clean up the question
        answer = find_answer(question, answers)
        questions.append({'q': question, 'a': answer})

        if answer:
            # the result to a reply will either contain another question or a confirmation
            result = reply_callback(first_data['id'], True)
        else:
            raise RuntimeError(
                f'A question was not given a positive reply. Question: "{question}". Answers: \n{pprint.pformat(answers)}\n. Request: {result.request}'
            )

    raise RuntimeError(f'Too many questions: {original_result}: {questions}')


@dataclass
class OrderRequest:
    conid: Optional[int]
    side: str
    quantity: float
    order_type: str
    acct_id: str

    # optional
    price: Optional[float] = field(default=None)
    conidex: Optional[str] = field(default=None)
    manual_indicator: Optional[bool] = field(default=None)
    ext_operator: Optional[str] = field(default=None)
    sec_type: Optional[str] = field(default=None)
    coid: Optional[str] = field(default=None)
    parent_id: Optional[str] = field(default=None)
    listing_exchange: Optional[str] = field(default=None)
    is_single_group: Optional[bool] = field(default=None)
    outside_rth: Optional[bool] = field(default=None)
    aux_price: Optional[float] = field(default=None)
    ticker: Optional[str] = field(default=None)
    tif: Optional[str] = field(default='GTC')
    trailing_amt: Optional[float] = field(default=None)
    trailing_type: Optional[str] = field(default=None)
    customer_account: Optional[str] = field(default=None)
    is_pro_customer: Optional[bool] = field(default=None)
    referrer: Optional[str] = field(default=None)
    cash_qty: Optional[float] = field(default=None)
    fx_qty: Optional[float] = field(default=None)
    use_adaptive: Optional[bool] = field(default=None)
    is_ccy_conv: Optional[bool] = field(default=None)
    allocation_method: Optional[str] = field(default=None)
    manual_order_time: Optional[int] = field(default=None)
    deactivated: Optional[bool] = field(default=None)
    strategy: Optional[str] = field(default=None)
    strategy_parameters: Optional[dict] = field(default=None)

    # undocumented
    is_close: Optional[bool] = field(default=None)

    # any additional fields that have not yet been mapped
    custom_fields: Optional[dict] = field(default=None)

    def to_dict(self) -> dict:
        """Convert dataclass to a dictionary, excluding None values."""
        return {f.name: getattr(self, f.name) for f in fields(self) if getattr(self, f.name) is not None}


_ORDER_REQUEST_MAPPING = {
    'conid': 'conid',
    'side': 'side',
    'quantity': 'quantity',
    'order_type': 'orderType',
    'price': 'price',
    'coid': 'cOID',
    'acct_id': 'acctId',
    'conidex': 'conidex',
    'manual_indicator': 'manualIndicator',
    'ext_operator': 'extOperator',
    'sec_type': 'secType',
    'parent_id': 'parentId',
    'listing_exchange': 'listingExchange',
    'is_single_group': 'isSingleGroup',
    'outside_rth': 'outsideRTH',
    'aux_price': 'auxPrice',
    'ticker': 'ticker',
    'tif': 'tif',
    'trailing_amt': 'trailingAmt',
    'trailing_type': 'trailingType',
    'customer_account': 'customerAccount',
    'is_pro_customer': 'isProCustomer',
    'referrer': 'referrer',
    'cash_qty': 'cashQty',
    'fx_qty': 'fxQty',
    'use_adaptive': 'useAdaptive',
    'is_ccy_conv': 'isCcyConv',
    'allocation_method': 'allocationMethod',
    'manual_order_time': 'manualOrderTime',
    'deactivated': 'deactivated',
    'strategy': 'strategy',
    'strategy_parameters': 'strategyParameters',
    'is_close': 'isClose',
}


def parse_order_request(order_request: OrderRequest, mapping: dict = None) -> dict:
    if mapping is None:
        mapping = _ORDER_REQUEST_MAPPING

    if isinstance(order_request, dict):
        _LOGGER.warning("Order request supplied as a dict. Use 'OrderRequest' dataclass instead.")
        d = order_request
    else:
        d = {mapping[k]: v for k, v in order_request.to_dict().items() if v is not None and k != 'custom_fields'}

        if order_request.custom_fields is not None:
            d.update(order_request.custom_fields)

    if 'conidex' in d and 'conid' in d:
        raise ValueError("Both 'conidex' and 'conid' are provided. When using 'conidex', specify `conid=None`.")

    return d


def make_order_request(
    conid: Union[int, str],
    side: str,
    quantity: float,
    order_type: str,
    acct_id: str,
    # optional
    price: float = None,
    conidex: str = None,
    sec_type: str = None,
    coid: str = None,
    parent_id: str = None,
    listing_exchange: str = None,
    is_single_group: bool = None,
    outside_rth: bool = None,
    aux_price: float = None,
    ticker: str = None,
    tif: str = 'GTC',
    trailing_amt: float = None,
    trailing_type: str = None,
    referrer: str = None,
    cash_qty: float = None,
    fx_qty: float = None,
    use_adaptive: bool = None,
    is_ccy_conv: bool = None,
    allocation_method: str = None,
    strategy: str = None,
    strategy_parameters=None,
):  # pragma: no cover
    """
     Create an order request object. Arguments set as None will not be included.

    Parameters:
        conid (int | str): Identifier of the security to trade.
        side (str): Order side, either 'SELL' or 'BUY'.
        quantity (int): Order quantity in number of shares.
        order_type (str): Type of the order (e.g., LMT, MKT, STP).
        price (float): Order limit price, depends on order type.
        coid (str): Customer Order ID, unique for a 24h span.
        acctId (str, optional): Account ID, defaults to the first account if not provided.

        conidex (str, Optional): Concatenated value of contract identifier and exchange.
        sec_type (str, Optional): Concatenated value of contract-identifier and security type.
        parent_id (str, Optional): Used for child orders in bracket orders, must match the parent's cOID.
        listing_exchange (str, Optional, optional): Exchange for order routing, default is "SMART".
        is_single_group (bool, Optional): Set to True for placing single group orders (OCA).
        outside_rth (bool, Optional): Set to True if the order can be executed outside regular trading hours.
        aux_price (float, Optional): Auxiliary price parameter.
        ticker (str, Optional): Underlying symbol for the contract.
        tif (str, Optional): Time-In-Force for the order (e.g., GTC, OPG, DAY, IOC). Default: "GTC".
        trailing_amt (float, Optional): Trailing amount for TRAIL or TRAILLMT orders.
        trailing_type (str, Optional): Trailing type ('amt' or '%') for TRAIL or TRAILLMT orders.
        referrer (str, Optional): Custom order reference.
        cash_qty (float, Optional): Cash Quantity for the order.
        fx_qty (float, Optional): Cash quantity for Currency Conversion Orders.
        use_adaptive (bool, Optional): Set to True to use the Price Management Algo.
        is_ccy_conv (bool, Optional): Set to True for FX conversion orders.
        allocation_method (str, Optional): Allocation method for FA account orders.
        strategy (str, Optional): IB Algo algorithm to use for the order.
        strategy_parameters (dict, Optional): Parameters for the specified IB Algo algorithm.
    """
    warn("'make_order_request' is deprecated. Use 'OrderRequest' dataclass instead.", DeprecationWarning, stacklevel=2)

    order_request = {}

    if conid is not None:
        order_request['conid'] = int(conid)

    if side is not None:
        order_request['side'] = str(side)

    if quantity is not None:
        order_request['quantity'] = int(quantity)

    if order_type is not None:
        order_request['orderType'] = str(order_type)

    if price is not None:
        order_request['price'] = price

    if coid is not None:
        order_request['cOID'] = coid

    # optional

    if acct_id is not None:
        order_request['acctId'] = acct_id

    if conidex is not None:
        order_request['conidex'] = conidex

    if sec_type is not None:
        order_request['secType'] = sec_type

    if parent_id is not None:
        order_request['parentId'] = parent_id

    if listing_exchange is not None:
        order_request['listingExchange'] = listing_exchange

    if is_single_group is not None:
        order_request['isSingleGroup'] = is_single_group

    if outside_rth is not None:
        order_request['outsideRTH'] = outside_rth

    if aux_price is not None:
        order_request['auxPrice'] = aux_price

    if ticker is not None:
        order_request['ticker'] = ticker

    if tif is not None:
        order_request['tif'] = tif

    if trailing_amt is not None:
        order_request['trailingAmt'] = trailing_amt

    if trailing_type is not None:
        order_request['trailingType'] = trailing_type

    if referrer is not None:
        order_request['referrer'] = referrer

    if cash_qty is not None:
        order_request['cashQty'] = cash_qty

    if fx_qty is not None:
        order_request['fxQty'] = fx_qty

    if use_adaptive is not None:
        order_request['useAdaptive'] = use_adaptive

    if is_ccy_conv is not None:
        order_request['isCcyConv'] = is_ccy_conv

    if allocation_method is not None:
        order_request['allocationMethod'] = allocation_method

    if strategy is not None:
        order_request['strategy'] = strategy

    if strategy_parameters is not None:
        order_request['strategyParameters'] = strategy_parameters

    return order_request


def date_from_ibkr(d: str) -> datetime.datetime:
    try:
        return datetime.datetime(int(d[:4]), int(d[4:6]), int(d[6:8]), int(d[8:10]), int(d[10:12]), int(d[12:14]))
    except ValueError:
        raise ValueError(
            f'Date seems to be missing fields: year={d[0:4]}, month={d[4:6]}, day={d[6:8]}, hour={d[8:10]}, minute={d[10:12]}, second={d[12:14]}'
        )


def extract_conid(data):
    # by default conid should be made available as 'smh+<conid>', let's look for it
    if 'topic' in data and '+' in data['topic']:
        return data['topic'].split('+')[-1]

    # as a backup we try to include the conid in the payload, IBKR seems to send it back to us
    elif 'payload' in data and 'conid' in data['payload']:
        return data['payload']['conid']

    return None


class Tickler:
    """
    Utility class used for maintaining the OAuth connection alive by repeatedly calling the `tickle` method.

    The Tickler runs in a separate thread and periodically sends requests to the IBKR API to prevent
    the session from expiring. This is essential for keeping the OAuth session active.
    """

    def __init__(self, client: 'IbkrClient', interval: Union[int, float] = var.IBIND_TICKLER_INTERVAL):
        """
        Initializes the Tickler instance.

        Parameters:
            client (IbkrClient): The client instance with a `tickle` method that maintains the session.
            interval (Union[int, float]): Interval between tickles in seconds. Default is 60 seconds.
        """
        self._client = client
        self._interval = interval
        self._stop_event = threading.Event()
        self._thread = None

    def _worker(self):
        _LOGGER.info(f'Tickler starts with interval={self._interval} seconds.')
        while not self._stop_event.wait(self._interval):
            try:
                self._client.tickle()
            except KeyboardInterrupt:
                _LOGGER.info('Tickler interrupted')
                break
            except TimeoutError:
                _LOGGER.warning('Tickler encountered a timeout error. This could indicate the servers are restarting. Investigate further if you see this log repeat frequently.')
            except Exception as e:
                _LOGGER.error(f'Tickler error: {exception_to_string(e)}')

        _LOGGER.info('Tickler gracefully stopped.')

    def start(self):
        """
        Starts the Tickler in a separate thread.

        This method creates and starts a daemon thread that periodically calls `tickle()` to keep the
        session alive.
        """
        if self._thread is not None:
            _LOGGER.info('Tickler thread already running. Stop the existing thread first by calling Tickler.stop()')
            return

        self._stop_event.clear()
        self._thread = threading.Thread(target=self._worker, daemon=True)
        self._thread.start()

    def stop(self, timeout:float=None):
        """
        Stops the Tickler thread.

        This method stops and joins the Tickler thread.

        Parameters:
            timeout (Optional[float]): Maximum time to wait for the Tickler thread to terminate.
                                       If None, waits indefinitely.
        """
        if self._thread is None:
            return

        self._stop_event.set()  # Wake up the sleeping thread immediately
        self._thread.join(timeout)
        self._thread = None  # Ensure cleanup


def cleanup_market_history_responses(
    market_history_response: Dict[str, Union[Exception, Result]],
    raise_on_error: bool = False,
):
    """
    Processes and cleans up market history responses, converting raw data into structured records.

    This function iterates over market history responses, extracts relevant trading data, and
    formats it into a structured dictionary. If errors are encountered, they are either logged
    and included in the results or raised based on the `raise_on_error` flag.

    Parameters:
        market_history_response (Dict[str, Union[Exception, Result]]):
            A dictionary where keys are symbols and values are either:
            - `Result` objects containing market history data.
            - `Exception` instances representing failed requests.

        raise_on_error (bool, optional):
            If `True`, raises encountered exceptions instead of storing them in the results.
            Defaults to `False`.

    Returns:
        Dict[str, Union[list[dict], Exception]]:
            A dictionary where keys are symbols and values are either:
            - A list of structured historical market data records, each containing:
                - `"open"`: Open price.
                - `"high"`: High price.
                - `"low"`: Low price.
                - `"close"`: Close price.
                - `"volume"`: Trading volume.
                - `"date"`: Datetime object representing the timestamp.
            - An `Exception` object if an error occurred (only if `raise_on_error=False`).

    Logs:
        - Errors when fetching market data if `raise_on_error=True`.
        - Warnings if market data is not live (i.e., missing 'S' or 'R' in `mdAvailability`).

    Raises:
        Exception: If `raise_on_error=True` and an error occurs.

    Example:
        >>> market_history_responses = {
        ...     symbol: client.marketdata_history_by_conid(**request)
        ...     for symbol, request in requests.items()
        ... }  # fmt: skip
        >>> results = cleanup_market_history_responses(market_history_responses)
        {
            "AAPL": [
                {"open": 150.0, "high": 155.0, "low": 149.0, "close": 153.0,
                 "volume": 500000, "date": datetime.datetime(2023, 11, 14, 10, 53, 20)}
            ],
            "TSLA": Exception("Failed to fetch data")
        }
    """
    results = {}
    for symbol, entry in market_history_response.items():
        if isinstance(entry, Exception):  # pragma: no cover
            if '"error":"No data."' in str(entry):
                _LOGGER.info(f'"No data" error for {symbol} when fetching market data. Consider increasing period and bar sizes.')

            if raise_on_error:
                _LOGGER.error(f'Error fetching market data for {symbol}')
                raise entry
            else:
                results[symbol] = entry
                continue

        # check if entry['mdAvailability'] has 'S' or 'R' in it
        if 'mdAvailability' in entry.data and not (any((key in entry.data['mdAvailability'].upper()) for key in ['S', 'R'])):
            _LOGGER.warning(f'Market data for {symbol} is not live: {decode_data_availability(entry.data["mdAvailability"])}')

        data = entry.data['data']
        records = []
        for record in data:
            records.append(
                {
                    'open': record['o'],
                    'high': record['h'],
                    'low': record['l'],
                    'close': record['c'],
                    'volume': record['v'],
                    'date': datetime.datetime.fromtimestamp(record['t'] / 1000, tz=datetime.UTC),
                }
            )
        results[symbol] = records
    return results


responses = {
    'accounts': [
        {
            'PrepaidCrypto-P': False,
            'PrepaidCrypto-Z': False,
            'accountAlias': None,
            'accountId': 'DUMMY_ACCOUNT_ID123',
            'accountStatus': 1678852800000,
            'accountTitle': 'Geralt of Rivia',
            'accountVan': 'DUMMY_ACCOUNT_ID123',
            'brokerageAccess': True,
            'businessType': 'INDEPENDENT',
            'clearingStatus': 'O',
            'covestor': False,
            'currency': 'USD',
            'desc': 'DUMMY_ACCOUNT_ID123',
            'displayName': 'Geralt of Rivia',
            'faclient': False,
            'ibEntity': 'IBLLC-US',
            'id': 'DUMMY_ACCOUNT_ID123',
            'noClientTrading': False,
            'parent': {
                'accountId': '',
                'isMChild': False,
                'isMParent': False,
                'isMultiplex': False,
                'mmc': [],
            },
            'trackVirtualFXPortfolio': False,
            'tradingType': 'STKNOPT',
            'type': 'DEMO',
        }
    ],
    'auth_status': {
        'MAC': 'F4:03:43:CD:BD:C0',
        'authenticated': True,
        'competing': False,
        'connected': True,
        'fail': '',
        'hardware_info': '1c4558e8|F4:03:43:CD:BD:C0',
        'message': '',
        'serverInfo': {
            'serverName': 'JifC11113',
            'serverVersion': 'Build 10.25.0m, Nov 17, 2023 11:29:27 AM',
        },
    },
    'conids': {
        'AAPL': 265598,
        'BBVA': 4815,
        'CDN': 453310666,
        'CFC': 42001300,
        'GOOG': 208813720,
        'HUBS': 169544810,
        'META': 107113386,
        'MSFT': 272093,
        'SAN': 12442,
        'SCHW': 11905,
        'TEAM': 589316251,
    },
    'filtered_conids': {
        'AAPL': 532640894,
        'BBVA': 4815,
        'CDN': 453310666,
        'CFC': 42001300,
        'GOOG': 210810667,
        'HUBS': 169544810,
        'META': 114922621,
        'MSFT': 272093,
        'SAN': 12442,
        'SCHW': 11905,
        'TEAM': 589316251,
    },
    'history': {
        'AAPL': {
            'barLength': 2678400,
            'chartAnnotations': 'x/D 0.24/1440;',
            'data': [{'c': 189.37, 'h': 192.93, 'l': 170.12, 'o': 171.05, 't': 1698796800000, 'v': 6942998.04}],
            'high': '19293/6942998.04/1440',
            'low': '17012/6942998.04/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195436',
            'startTime': '20231031-00:00:00',
            'symbol': 'AAPL',
            'text': 'APPLE INC',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1,
        },
        'BBVA': {
            'barLength': 2678400,
            'data': [{'c': 9.45, 'h': 9.52, 'l': 7.94, 'o': 8.0, 't': 1698796800000, 'v': 211434.01}],
            'high': '952/211434.01/1440',
            'low': '794/211434.01/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195431',
            'startTime': '20231031-00:00:00',
            'symbol': 'BBVA',
            'text': 'BANCO BILBAO VIZCAYA-SP ADR',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
        'CDN': {
            'barLength': 2678400,
            'data': [{'c': 0.32, 'h': 0.37, 'l': 0.26, 'o': 0.37, 't': 1698796800000, 'v': 83773}],
            'high': '74/83773/1440',
            'low': '52/83773/1440',
            'mdAvailability': 'D',
            'messageVersion': 2,
            'mktDataDelay': 900,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '3',
            'priceFactor': 200,
            'serverId': '195417',
            'startTime': '20231031-00:00:00',
            'symbol': 'CDN',
            'text': 'CDN MAVERICK CAPITAL CORP',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1,
        },
        'CFC': {
            'barLength': 2678400,
            'data': [{'c': 1.62, 'h': 1.91, 'l': 1.48, 'o': 1.63, 't': 1698796800000, 'v': 949972}],
            'high': '3820/949972/1440',
            'low': '2960/949972/1440',
            'mdAvailability': 'D',
            'messageVersion': 2,
            'mktDataDelay': 900,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '4',
            'priceFactor': 2000,
            'serverId': '195419',
            'startTime': '20231031-00:00:00',
            'symbol': 'CFC',
            'text': 'UET UNITED ELECTRONIC TECHNO',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
        'GOOG': {
            'barLength': 2678400,
            'data': [{'c': 136.4, 'h': 141.1, 'l': 124.92, 'o': 125.24, 't': 1698796800000, 'v': 2346378.23}],
            'high': '14110/2346378.23/1440',
            'low': '12492/2346378.23/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195424',
            'startTime': '20231031-00:00:00',
            'symbol': 'GOOG',
            'text': 'ALPHABET INC-CL C',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
        'HUBS': {
            'barLength': 2678400,
            'data': [{'c': 496.52, 'h': 499.15, 'l': 408.62, 'o': 422.06, 't': 1698796800000, 'v': 77847.62}],
            'high': '49915/77847.62/1440',
            'low': '40862/77847.62/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195440',
            'startTime': '20231031-00:00:00',
            'symbol': 'HUBS',
            'text': 'HUBSPOT INC',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1,
        },
        'META': {
            'barLength': 2678400,
            'data': [{'c': 332.2, 'h': 342.92, 'l': 301.85, 'o': 301.85, 't': 1698796800000, 'v': 1826033.84}],
            'high': '34292/1826033.84/1440',
            'low': '30185/1826033.84/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195438',
            'startTime': '20231031-00:00:00',
            'symbol': 'META',
            'text': 'META PLATFORMS INC-CLASS A',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
        'MSFT': {
            'barLength': 2678400,
            'chartAnnotations': 'x/D 0.75/1440;',
            'data': [{'c': 378.85, 'h': 384.3, 'l': 339.65, 'o': 339.79, 't': 1698796800000, 'v': 2886663.6}],
            'high': '38430/2886663.6/1440',
            'low': '33965/2886663.6/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195433',
            'startTime': '20231031-00:00:00',
            'symbol': 'MSFT',
            'text': 'MICROSOFT CORP',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
        'SAN': {
            'barLength': 2678400,
            'data': [{'c': 4.13, 'h': 4.15, 'l': 3.62, 'o': 3.66, 't': 1698796800000, 'v': 466540.33}],
            'high': '41500/466540.33/1440',
            'low': '36200/466540.33/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '4',
            'priceFactor': 10000,
            'serverId': '195429',
            'startTime': '20231031-00:00:00',
            'symbol': 'SAN',
            'text': 'BANCO SANTANDER SA-SPON ADR',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1,
        },
        'SCHW': {
            'barLength': 2678400,
            'chartAnnotations': 'x/D 0.25/1440;',
            'data': [{'c': 59.61, 'h': 59.94, 'l': 51.86, 'o': 52.06, 't': 1698796800000, 'v': 1299913.92}],
            'high': '5994/1299913.92/1440',
            'low': '5186/1299913.92/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195422',
            'startTime': '20231031-00:00:00',
            'symbol': 'SCHW',
            'text': 'SCHWAB (CHARLES) CORP',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
        'TEAM': {
            'barLength': 2678400,
            'data': [{'c': 195.55, 'h': 199.36, 'l': 165.51, 'o': 180.7, 't': 1698796800000, 'v': 171177.85}],
            'high': '19936/171177.85/1440',
            'low': '16551/171177.85/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195426',
            'startTime': '20231031-00:00:00',
            'symbol': 'TEAM',
            'text': 'ATLASSIAN CORP-CL A',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1,
        },
    },
    'ledger': {
        'BASE': {
            'acctcode': 'DUMMY_ACCOUNT_ID123',
            'cashbalance': 985376.94,
            'cashbalancefxsegment': 0.0,
            'commoditymarketvalue': 0.0,
            'corporatebondsmarketvalue': 0.0,
            'cryptocurrencyvalue': 0.0,
            'currency': 'BASE',
            'dividends': 0.0,
            'exchangerate': 1,
            'funds': 0.0,
            'futuremarketvalue': 0.0,
            'futureoptionmarketvalue': 0.0,
            'futuresonlypnl': 0.0,
            'interest': 3401.28,
            'issueroptionsmarketvalue': 0.0,
            'key': 'LedgerList',
            'moneyfunds': 0.0,
            'netliquidationvalue': 1034222.25,
            'realizedpnl': 0.0,
            'secondkey': 'BASE',
            'sessionid': 1,
            'settledcash': 985376.94,
            'severity': 0,
            'stockmarketvalue': 45444.0,
            'stockoptionmarketvalue': 0.0,
            'tbillsmarketvalue': 0.0,
            'tbondsmarketvalue': 0.0,
            'timestamp': 1701198108,
            'unrealizedpnl': 1701.0,
            'warrantsmarketvalue': 0.0,
        },
        'USD': {
            'acctcode': 'DUMMY_ACCOUNT_ID123',
            'cashbalance': 985376.94,
            'cashbalancefxsegment': 0.0,
            'commoditymarketvalue': 0.0,
            'corporatebondsmarketvalue': 0.0,
            'cryptocurrencyvalue': 0.0,
            'currency': 'USD',
            'dividends': 0.0,
            'endofbundle': 1,
            'exchangerate': 1,
            'funds': 0.0,
            'futuremarketvalue': 0.0,
            'futureoptionmarketvalue': 0.0,
            'futuresonlypnl': 0.0,
            'interest': 3401.28,
            'issueroptionsmarketvalue': 0.0,
            'key': 'LedgerList',
            'moneyfunds': 0.0,
            'netliquidationvalue': 1034222.25,
            'realizedpnl': 0.0,
            'secondkey': 'USD',
            'sessionid': 1,
            'settledcash': 985376.94,
            'severity': 0,
            'stockmarketvalue': 45444.0,
            'stockoptionmarketvalue': 0.0,
            'tbillsmarketvalue': 0.0,
            'tbondsmarketvalue': 0.0,
            'timestamp': 1701198108,
            'unrealizedpnl': 1701.0,
            'warrantsmarketvalue': 0.0,
        },
    },
    'positions': [
        {
            'acctId': 'DUMMY_ACCOUNT_ID123',
            'allExchanges': 'AMEX,NYSE,CBOE,PHLX,CHX,ARCA,ISLAND,ISE,IDEAL,NASDAQQ,DRCTEDGE,BEX,BATS,NITEECN,EDGEA,CSFBALGO,JEFFALGO,PSX,BYX,ITG,PDQ,IBKRATS,CITADEL,NYSEDARK,MIAX,IBDARK,CITADELDP,NASDDARK,IEX,WEDBUSH,SUMMER,WINSLOW,LIQITG,UBSDARK,BTIG,VIRTU,JEFF,OPCO,COWEN,DBK,JPMC,EDGX,JANE,NEEDHAM,FRACSHARE,RBCALGO,VIRTUDP,BAYCREST,FOXRIVER,MND,NITEEXST,PEARL,GSDARK,NITERTL,NYSENAT,IEXMID,HRT,FLOWTRADE,HRTDP,JANELP,PEAK6,IMCDP,CTDLZERO,HRTMID,JANEZERO,HRTEXST,IMCLP,LTSE,SOCGENDP,MEMX,INTELCROS,VIRTUBYIN,JUMPTRADE,NITEZERO,TPLUS1,XTXEXST,XTXDP,XTXMID,COWENLP,BARCDP,JUMPLP,OLDMCLP,RBCCMALP,WALLBETH,IBEOS,JONES,GSLP,BLUEOCEAN,USIBSILP,OVERNIGHT,JANEMID,IBATSEOS,HRTZERO',
            'assetClass': 'STK',
            'avgCost': 437.43,
            'avgPrice': 437.43,
            'chineseName': 'SPDR&#x6807;&#x666E;500 ETF&#x4FE1;&#x6258;',
            'conExchMap': [],
            'conid': 756733,
            'contractDesc': 'SPY',
            'countryCode': 'US',
            'currency': 'USD',
            'displayRule': {
                'displayRuleStep': [{'decimalDigits': 2, 'lowerEdge': 0.0, 'wholeDigits': 4}],
                'magnification': 0,
            },
            'exchs': None,
            'exerciseStyle': None,
            'expiry': None,
            'fullName': 'SPY',
            'group': None,
            'hasOptions': True,
            'incrementRules': [{'increment': 0.01, 'lowerEdge': 0.0}],
            'isEventContract': False,
            'isUS': True,
            'lastTradingDay': None,
            'listingExchange': 'ARCA',
            'mktPrice': 454.44000245,
            'mktValue': 45444.0,
            'model': '',
            'multiplier': 0.0,
            'name': 'SPDR S&P 500 ETF TRUST',
            'pageSize': 100,
            'position': 100.0,
            'putOrCall': None,
            'realizedPnl': 0.0,
            'sector': None,
            'sectorGroup': None,
            'strike': '0',
            'ticker': 'SPY',
            'time': 30,
            'type': 'ETF',
            'undConid': 0,
            'unrealizedPnl': 1701.0,
        }
    ],
    'stocks': {
        'AAPL': [
            {
                'assetClass': 'STK',
                'chineseName': '&#x82F9;&#x679C;&#x516C;&#x53F8;',
                'contracts': [
                    {'conid': 265598, 'exchange': 'NASDAQ', 'isUS': True},
                    {'conid': 38708077, 'exchange': 'MEXI', 'isUS': False},
                    {'conid': 273982664, 'exchange': 'EBS', 'isUS': False},
                ],
                'name': 'APPLE INC',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 493546048, 'exchange': 'LSEETF', 'isUS': False}],
                'name': 'LS 1X AAPL',
            },
            {
                'assetClass': 'STK',
                'chineseName': '&#x82F9;&#x679C;&#x516C;&#x53F8;',
                'contracts': [{'conid': 532640894, 'exchange': 'AEQLIT', 'isUS': False}],
                'name': 'APPLE INC-CDR',
            },
        ],
        'BBVA': [
            {
                'assetClass': 'STK',
                'chineseName': '&#x897F;&#x73ED;&#x7259;&#x5BF9;&#x5916;&#x94F6;&#x884C;',
                'contracts': [
                    {'conid': 30314149, 'exchange': 'BM', 'isUS': False},
                    {'conid': 38708792, 'exchange': 'MEXI', 'isUS': False},
                ],
                'name': 'BANCO BILBAO VIZCAYA ARGENTA',
            },
            {
                'assetClass': 'STK',
                'chineseName': '&#x897F;&#x73ED;&#x7259;&#x5BF9;&#x5916;&#x94F6;&#x884C;',
                'contracts': [{'conid': 4815, 'exchange': 'NYSE', 'isUS': True}],
                'name': 'BANCO BILBAO VIZCAYA-SP ADR',
            },
        ],
        'CDN': [
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 453310666, 'exchange': 'PURE', 'isUS': False}],
                'name': 'CDN MAVERICK CAPITAL CORP',
            }
        ],
        'CFC': [
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 42001300, 'exchange': 'IBIS', 'isUS': False}],
                'name': 'UET UNITED ELECTRONIC TECHNO',
            }
        ],
        'GOOG': [
            {
                'assetClass': 'STK',
                'chineseName': 'Alphabet&#x516C;&#x53F8;',
                'contracts': [
                    {'conid': 208813720, 'exchange': 'NASDAQ', 'isUS': True},
                    {'conid': 210810667, 'exchange': 'MEXI', 'isUS': False},
                ],
                'name': 'ALPHABET INC-CL C',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 493546022, 'exchange': 'LSEETF', 'isUS': False}],
                'name': 'LS 1X GOOG',
            },
            {
                'assetClass': 'STK',
                'chineseName': 'Alphabet&#x516C;&#x53F8;',
                'contracts': [{'conid': 532638805, 'exchange': 'AEQLIT', 'isUS': False}],
                'name': 'ALPHABET INC - CDR',
            },
        ],
        'HUBS': [
            {
                'assetClass': 'STK',
                'chineseName': 'HubSpot&#x516C;&#x53F8;',
                'contracts': [{'conid': 169544810, 'exchange': 'NYSE', 'isUS': True}],
                'name': 'HUBSPOT INC',
            }
        ],
        'META': [
            {
                'assetClass': 'STK',
                'chineseName': 'Meta&#x5E73;&#x53F0;&#x80A1;&#x4EFD;&#x6709;&#x9650;&#x516C;&#x53F8;',
                'contracts': [
                    {'conid': 107113386, 'exchange': 'NASDAQ', 'isUS': True},
                    {'conid': 114922621, 'exchange': 'MEXI', 'isUS': False},
                ],
                'name': 'META PLATFORMS INC-CLASS A',
            },
            {
                'assetClass': 'STK',
                'chineseName': 'WisdomTree&#x5DE5;&#x4E1A;&#x91D1;&#x5C5E;&#x589E;&#x5F3A;',
                'contracts': [
                    {'conid': 355191857, 'exchange': 'LSEETF', 'isUS': False},
                    {'conid': 355194893, 'exchange': 'BVME.ETF', 'isUS': False},
                ],
                'name': 'WT INDUS METALS ENHANCED',
            },
            {
                'assetClass': 'STK',
                'chineseName': 'Meta&#x5E73;&#x53F0;&#x80A1;&#x4EFD;&#x6709;&#x9650;&#x516C;&#x53F8;',
                'contracts': [{'conid': 530091499, 'exchange': 'AEQLIT', 'isUS': False}],
                'name': 'META PLATFORMS INC-CDR',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 600541860, 'exchange': 'SFB', 'isUS': False}],
                'name': 'METACON AB',
            },
        ],
        'MSFT': [
            {
                'assetClass': 'STK',
                'chineseName': '&#x5FAE;&#x8F6F;&#x516C;&#x53F8;',
                'contracts': [
                    {'conid': 272093, 'exchange': 'NASDAQ', 'isUS': True},
                    {'conid': 38708990, 'exchange': 'MEXI', 'isUS': False},
                    {'conid': 415569505, 'exchange': 'EBS', 'isUS': False},
                ],
                'name': 'MICROSOFT CORP',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 493546075, 'exchange': 'LSEETF', 'isUS': False}],
                'name': 'LS 1X MSFT',
            },
            {
                'assetClass': 'STK',
                'chineseName': '&#x5FAE;&#x8F6F;&#x516C;&#x53F8;',
                'contracts': [{'conid': 518938052, 'exchange': 'AEQLIT', 'isUS': False}],
                'name': 'MICROSOFT CORP-CDR',
            },
        ],
        'SAN': [
            {
                'assetClass': 'STK',
                'chineseName': '&#x6851;&#x5766;&#x5FB7;',
                'contracts': [
                    {'conid': 38708867, 'exchange': 'MEXI', 'isUS': False},
                    {'conid': 385055564, 'exchange': 'WSE', 'isUS': False},
                ],
                'name': 'BANCO SANTANDER SA',
            },
            {
                'assetClass': 'STK',
                'chineseName': '&#x8D5B;&#x8BFA;&#x83F2;',
                'contracts': [{'conid': 29612249, 'exchange': 'SBF', 'isUS': False}],
                'name': 'SANOFI',
            },
            {
                'assetClass': 'STK',
                'chineseName': '&#x6851;&#x5766;&#x5FB7;',
                'contracts': [{'conid': 12442, 'exchange': 'NYSE', 'isUS': True}],
                'name': 'BANCO SANTANDER SA-SPON ADR',
            },
            {
                'assetClass': 'STK',
                'chineseName': '&#x6851;&#x5766;&#x5FB7;&#x82F1;&#x56FD;&#x516C;&#x5171;&#x6709;&#x9650;&#x516C;&#x53F8;',
                'contracts': [{'conid': 80993135, 'exchange': 'LSE', 'isUS': False}],
                'name': 'SANTANDER UK PLC',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 199284645, 'exchange': 'ASX', 'isUS': False}],
                'name': 'SAGALIO ENERGY LTD',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 221033380, 'exchange': 'VALUE', 'isUS': False}],
                'name': 'CONCORD GREEN ENERGY HOLDINGS',
            },
        ],
        'SCHW': [
            {
                'assetClass': 'STK',
                'chineseName': '&#x5609;&#x4FE1;&#x7406;&#x8D22;',
                'contracts': [
                    {'conid': 11905, 'exchange': 'NYSE', 'isUS': True},
                    {'conid': 174320486, 'exchange': 'MEXI', 'isUS': False},
                ],
                'name': 'SCHWAB (CHARLES) CORP',
            }
        ],
        'TEAM': [
            {
                'assetClass': 'STK',
                'chineseName': 'Team Tankers&#x56FD;&#x9645;&#x6709;&#x9650;&#x516C;&#x53F8;',
                'contracts': [{'conid': 251279169, 'exchange': 'VALUE', 'isUS': False}],
                'name': 'TEAM TANKERS INTERNATIONAL L',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 475655801, 'exchange': 'LSE', 'isUS': False}],
                'name': 'TEAM PLC',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 589316251, 'exchange': 'NASDAQ', 'isUS': True}],
                'name': 'ATLASSIAN CORP-CL A',
            },
            {
                'assetClass': 'STK',
                'chineseName': None,
                'contracts': [{'conid': 638360210, 'exchange': 'PURE', 'isUS': False}],
                'name': 'CANADIAN NEXUS TEAM VENTURES',
            },
        ],
    },
    'tickle': {
        'collission': False,
        'hmds': {'error': 'no bridge'},
        'iserver': {
            'authStatus': {
                'MAC': 'F4:03:43:CD:BD:C0',
                'authenticated': True,
                'competing': False,
                'connected': True,
                'hardware_info': '1c4558e8|F4:03:43:CD:BD:C0',
                'message': '',
                'serverInfo': {
                    'serverName': 'JifC11113',
                    'serverVersion': 'Build 10.25.0m, Nov 17, 2023 11:29:27 AM',
                },
            }
        },
        'session': 'ce109cbeb462a76dc3f6f4182c830cc1',
        'ssoExpires': 585695,
        'userId': 99996016,
    },
    'live_orders': [
        {
            'account': 'DUMMY_ACCOUNT_ID123',
            'acct': 'DUMMY_ACCOUNT_ID123',
            'bgColor': '#FFFFFF',
            'cashCcy': 'USD',
            'companyName': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'description1': 'SPY',
            'fgColor': '#0000CC',
            'filledQuantity': 0.0,
            'lastExecutionTime': '240103201041',
            'lastExecutionTime_r': 1704312641000,
            'listingExchange': 'ARCA',
            'orderDesc': 'Buy 1 Limit 457.00, GTC',
            'orderId': 1210386759,
            'orderType': 'Limit',
            'order_ccp_status': 'Submitted',
            'order_ref': '240103-SPY-1-BUY-L@457-bxv8',
            'origOrderType': 'LIMIT',
            'outsideRTH': False,
            'price': '457.00',
            'remainingQuantity': 1.0,
            'secType': 'STK',
            'side': 'BUY',
            'sizeAndFills': '0/1',
            'status': 'PreSubmitted',
            'supportsTaxOpt': '1',
            'ticker': 'SPY',
            'timeInForce': 'GTC',
            'totalSize': 1.0,
        },
        {
            'account': 'DUMMY_ACCOUNT_ID123',
            'acct': 'DUMMY_ACCOUNT_ID123',
            'bgColor': '#FFFFFF',
            'cashCcy': 'USD',
            'companyName': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'description1': 'SPY',
            'fgColor': '#0000CC',
            'filledQuantity': 0.0,
            'lastExecutionTime': '240103200639',
            'lastExecutionTime_r': 1704312399000,
            'listingExchange': 'ARCA',
            'orderDesc': 'Buy 1 Limit 457.00, GTC',
            'orderId': 1210386726,
            'orderType': 'Limit',
            'order_ccp_status': 'Submitted',
            'order_ref': '240103-SPY-1-BUY-L@457-pfqd',
            'origOrderType': 'LIMIT',
            'outsideRTH': False,
            'price': '457.00',
            'remainingQuantity': 1.0,
            'secType': 'STK',
            'side': 'BUY',
            'sizeAndFills': '0/1',
            'status': 'PreSubmitted',
            'supportsTaxOpt': '1',
            'ticker': 'SPY',
            'timeInForce': 'GTC',
            'totalSize': 1.0,
        },
        {
            'account': 'DUMMY_ACCOUNT_ID123',
            'acct': 'DUMMY_ACCOUNT_ID123',
            'bgColor': '#FFFFFF',
            'cashCcy': 'USD',
            'companyName': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'description1': 'SPY',
            'fgColor': '#0000CC',
            'filledQuantity': 0.0,
            'lastExecutionTime': '240103010227',
            'lastExecutionTime_r': 1704243747000,
            'listingExchange': 'ARCA',
            'orderDesc': 'Buy 1 Limit 453.00, GTC',
            'orderId': 1928088808,
            'orderType': 'Limit',
            'order_ccp_status': 'Submitted',
            'order_ref': '240103-SPY-1-BUY-L@453-v5DU',
            'origOrderType': 'LIMIT',
            'outsideRTH': False,
            'price': '453.00',
            'remainingQuantity': 1.0,
            'secType': 'STK',
            'side': 'BUY',
            'sizeAndFills': '0/1',
            'status': 'PreSubmitted',
            'supportsTaxOpt': '1',
            'ticker': 'SPY',
            'timeInForce': 'GTC',
            'totalSize': 1.0,
        },
        {
            'account': 'DUMMY_ACCOUNT_ID123',
            'acct': 'DUMMY_ACCOUNT_ID123',
            'bgColor': '#FFFFFF',
            'cashCcy': 'USD',
            'companyName': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'description1': 'SPY',
            'fgColor': '#0000CC',
            'filledQuantity': 0.0,
            'lastExecutionTime': '240103212831',
            'lastExecutionTime_r': 1704317311000,
            'listingExchange': 'ARCA',
            'orderDesc': 'Buy 1 Limit 458.00, GTC',
            'orderId': 1210387040,
            'orderType': 'Limit',
            'order_ccp_status': 'Submitted',
            'order_ref': '240103-SPY-1-BUY-L@458-J5cN',
            'origOrderType': 'LIMIT',
            'outsideRTH': False,
            'price': '458.00',
            'remainingQuantity': 1.0,
            'secType': 'STK',
            'side': 'BUY',
            'sizeAndFills': '0/1',
            'status': 'PreSubmitted',
            'supportsTaxOpt': '1',
            'ticker': 'SPY',
            'timeInForce': 'GTC',
            'totalSize': 1.0,
        },
    ],
    'order_status': {
        '1210386759': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#0000CC',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 457.00, GTC',
            'exit_strategy_display_price': '457.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#FFFFFF',
            'limit_price': '457.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 457.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 457.00, GTC',
            'order_id': 1210386759,
            'order_not_editable': False,
            'order_status': 'PreSubmitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103201041',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '37888',
            'sec_type': 'STK',
            'server_id': '0',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False,
        },
        '1210386726': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#0000CC',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 457.00, GTC',
            'exit_strategy_display_price': '457.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#FFFFFF',
            'limit_price': '457.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 457.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 457.00, GTC',
            'order_id': 1210386726,
            'order_not_editable': False,
            'order_status': 'PreSubmitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103200639',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '37889',
            'sec_type': 'STK',
            'server_id': '1',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False,
        },
        '1928088808': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#0000CC',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 453.00, GTC',
            'exit_strategy_display_price': '453.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#FFFFFF',
            'limit_price': '453.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 453.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 453.00, GTC',
            'order_id': 1928088808,
            'order_not_editable': False,
            'order_status': 'PreSubmitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103010227',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '37890',
            'sec_type': 'STK',
            'server_id': '2',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False,
        },
        '1210386714': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#00F000',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 457.00, GTC',
            'exit_strategy_display_price': '457.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#000000',
            'limit_price': '457.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 457.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 457.00, GTC',
            'order_id': 1210386714,
            'order_not_editable': False,
            'order_status': 'Submitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103200503',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '216243',
            'sec_type': 'STK',
            'server_id': '824',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False,
        },
    },
}



from distutils.core import setup

from setuptools import find_packages


setup(name='ibind', packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests', 'examples', 'docs', 'out', 'dist']))
